/*
    mtrd  --  a network diagnostic tool
    Copyright (C) 1997,1998  Matt Kimball
    Copyright (C) 2000,2001  Christoph Lameter

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
   1999-08-13 ok Olav@okvittem.priv.no  added -psize

*/

#include <config.h>

#if defined(HAVE_SYS_XTI_H)
#include <sys/xti.h>
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <memory.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>

#include "net.h"
#include "output.h"


extern int Verbose;
extern double FloodDelay;

#define MaxTransit 4

/*  We can't rely on header files to provide this information, because
    the fields have different names between, for instance, Linux and 
    Solaris  */
struct ICMPHeader {
  uint8 type;
  uint8 code;
  uint16 checksum;
  uint16 id;
  uint16 sequence;
};

/*  Structure of an IP header.  */
struct IPHeader {
  uint8 version;
  uint8 tos;
  uint16 len;
  uint16 id;
  uint16 frag;
  uint8 ttl;
  uint8 protocol;
  uint16 check;
  uint32 saddr;
  uint32 daddr;
};
  
#define ICMP_ECHO		8
#define ICMP_ECHOREPLY		0

#define ICMP_TSTAMP		13
#define ICMP_TSTAMPREPLY	14

#define ICMP_TIME_EXCEEDED	11

#ifndef SOL_IP
#define SOL_IP 0
#endif


/* Memory requirements for a host = 40 bytes */
struct nethost {
  uint32 addr;
  int xmit;
  int returned;
  long long total;
  int best;
  int worst;
  double time;
  struct nethost *next;
};

struct sequence {
    struct netpath *p;	/* Which path are we are on */
    struct nethost *h;	/* Which host was this packet destined for ? */
    int transit;
    double time;
};


/* Configuration parameter: How many queries to unknown hosts do we
   send? (This limits the amount of traffic generated if a host is not
   reachable) */
#define MAX_UNKNOWN_HOSTS 5


static struct nethost **host;
static struct sequence sequence[MaxSequence];

#define HASH(x) ((uint32)(x) % MaxHost)

int sendsock;
int recvsock;
struct sockaddr_in remoteaddress;

int pid;

extern int packetsize;

int MaxPathList = 1024; /* This is going to be "amended" to larger sizes if necessary */

/* List of paths that we are monitoring size= 140 bytes*/
struct netpath {
	struct nethost*host;
	int counter;
	struct netpath *oldpath;	/* If !=NULL means that an old path exists that has to be written out as soon as new path has been completely determined */
	int n;				/* rotating pointer in path */
	struct nethost *path[MaxPathLength];	/* pointers to hosts we got responses from */
};

struct netpath **hpath;

int hosts;	/* Number of Hosts */
int hpaths;	/* Number of Paths */

int cpath;	/* Active path */

struct netpath *current;	/* Pointer to current path structure */

int net_init(void)  {
	hpath=calloc(MaxPathList,sizeof(struct netpath *));
	host=calloc(MaxHost,sizeof(struct nethost *));
	hosts=0;
	hpaths=0;
	cpath=0;
	current=NULL;
	if (hpath==NULL || host==NULL) return -1;
	sendsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
  	if(sendsock < 0) return -1;
	recvsock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if(recvsock < 0) return -1;
	fcntl(recvsock,F_SETFL,O_NONBLOCK);
	pid=getpid();
	return 0;
}

struct nethost * find_host(uint32 addr) {
   struct nethost *x;
   /* Simplified Hash. Uses prime in MaxHost */
   x=host[HASH(addr)];
   while (x) {
	if (x->addr==addr) return x;
	x=x->next;
   };
   return NULL;
}

struct nethost* add_host(uint32 addr) {
  struct nethost *h;
  uint32 v;
  if (addr==0) fatal("addhost: IP address zero!");
  if (find_host(addr)) fatal("addhost: duplicate IP address\n");
  h=calloc(1,sizeof(struct nethost));
  if (h==NULL) fatal("add_host: calloc failed\n");
  h->addr=addr;
  v=HASH(addr);
  h->next=host[v];
  host[v]=h;
  hosts++;
  if (find_host(addr)!=h) fatal("Consistency check failed in add_host()\n");
  return h;
}

void net_add_netpath(uint32 addr) {
    if (find_host(addr)) {
	output("net_add_netpath: duplicate IP %s",addr_to_str(addr));
	return;
    }
    if (hpaths>=MaxPathList) {
        struct netpath **new;
        MaxPathList = 2*MaxPathList;
        if ((new=calloc(MaxPathList,sizeof(struct netpath *)))==NULL) fatal("net_add_netpath: calloc failed\n");
        memcpy(new,hpath,MaxPathList*sizeof(struct netpath *)/2);
        free(hpath);
        hpath=new;
    }
    hpath[hpaths]=calloc(1,sizeof(struct netpath));
    hpath[hpaths++]->host=add_host(addr);
}

/* This doesn't work for odd sz. I don't know enough about this to say
   that this is wrong. It doesn't seem to cripple mtr though. -- REW */
int checksum(void *data, int sz) {
  unsigned short *ch;
  unsigned int sum;

  sum = 0;
  ch = data;
  sz = sz / 2;
  while(sz--) {
    sum += *(ch++);
  }
  
  sum = (sum >> 16) + (sum & 0xffff);  

  return (~sum & 0xffff);  
}


static int BSDfix = 0;

int new_sequence(struct nethost *q) {
  static int next_sequence = 0;
  int seq;

  seq = next_sequence++;
  if(next_sequence >= MaxSequence)
    next_sequence = 0;
    
  sequence[seq].p = current;
  sequence[seq].transit = 1;
  sequence[seq].h = q;
  sequence[seq].time=now();
  if (q) {
     if (q->xmit==0) q->time=now();
     q->xmit++;
  }
  return seq;
}

/*  Attempt to find the host at a particular number of hops away  */
void net_send_query(struct nethost *q) {
  char packet[MAXPACKET];
  struct IPHeader *ip;
  struct ICMPHeader *icmp;

  int rv;
  static int first=1;

  if ( packetsize < MINPACKET ) packetsize = MINPACKET;
  if ( packetsize > MAXPACKET ) packetsize = MAXPACKET;
  memset(packet, 0, packetsize);

  ip = (struct IPHeader *)packet;
  icmp = (struct ICMPHeader *)(packet + sizeof(struct IPHeader));

  ip->version = 0x45;
  ip->tos = 0;
  ip->len = BSDfix ? packetsize: htons (packetsize);
  ip->id = 0;
  ip->frag = 0;
  ip->ttl = current->n + 1;
  ip->protocol = IPPROTO_ICMP;
  ip->saddr = 0;

  remoteaddress.sin_family = AF_INET;
  ip->daddr = remoteaddress.sin_addr.s_addr = current->host->addr;

  icmp->type = ICMP_ECHO;
  /* This insures that we can categorize replies correctly */
  icmp->id = pid+current->n;
  icmp->sequence = new_sequence(q);

  icmp->checksum = checksum(icmp, packetsize - sizeof(struct IPHeader));
  ip->check = checksum(ip, packetsize);

  if (Verbose) output("ICMP to %s TTL=%d",addr_to_str(ip->daddr),ip->ttl);
  rv = sendto(sendsock, packet, packetsize, 0, 
	 (struct sockaddr *)&remoteaddress, sizeof(remoteaddress));
  if (first && (rv < 0) && (errno == EINVAL)) {
    ip->len = packetsize;
    rv = sendto(sendsock, packet, packetsize, 0, 
		(struct sockaddr *)&remoteaddress, sizeof(remoteaddress));
    if (rv >= 0) {
      output ("You've got a broken (FreeBSD?) system");
      BSDfix = 1;
    }
  }
  first = 0;
}

/* The following function dumps the data out to a file. Eventually it could be considered to
  spawn for doing that since the writing could take valuable time away from scanning */
extern char *datapath;
extern int format;

void freepaths(struct netpath *p) {
    if (p) {
       freepaths(p->oldpath);
       free(p);
    }
}

void data_dump(void) {
  FILE *f;
  int elements=0;
  char buffer[100];
  long rtime;long now=time(NULL);
  sprintf(buffer,"%s/path.%ld",datapath,now);
  output("Dumping Routing Path Data to %d hosts into %s",hpaths,buffer);
  f=fopen(buffer,"a");
  if (f==NULL) { perror(buffer); return;}
  { struct netpath *q;int i;int n;
     for(n=0;n<hpaths;n++) {
        int pathnr=0;
        for(q=hpath[n];q;q=q->oldpath,pathnr++) { 
            int hops=MaxPathLength;
            while (hops>0 && q->path[hops-1]==NULL) hops--;
            if (format==1)
            	fprintf(f,"%s\t%d\t%d",addr_to_str(q->host->addr),q->counter,hops);
	    if (q->counter) {
		for (i=0;i<hops;i++) {
		   if (q->path[i]) {
		      if (format==0)
			fprintf(f,"%lu\t%lu\t%u\t%u\t%lu\t%u\n",
			   htonl(q->host->addr),
			   now,
			   pathnr,
			   i+1,
			   htonl(q->path[i]->addr),
			   q->counter);
		      else
			fprintf(f,"\t%s",addr_to_str(q->path[i]->addr));
		    } else {
			if (format) fprintf(f,"\tNULL");
		  }
		  elements++;
		}
	    }
	    if (format) fprintf(f,"\n");
	    q->counter=0;
	}
	q=hpath[n];
	freepaths(q->oldpath);
	q->oldpath=NULL;
     }
  }
  fclose(f);
  output("%d routing elements in %d sec",elements,time(NULL)-now);
  action("dump-paths",buffer);
  sprintf(buffer,"%s/router.%ld",datapath,now);
  output("Dumping Router Data for %d routers into %s",hosts,buffer);
  f=fopen(buffer,"a");
  if (f==NULL) { perror(buffer); return;}
  elements=0;rtime=time(NULL);
  { struct nethost **p,*h;
     for(p=host;p<host+MaxHost;p++) {
        h=*p;
        while (h && h->xmit > 0) {
            if (format==0)
             fprintf(f,"%u\t%lu\t%u\t%u\t%u\t%u\t%u\n",
	      htonl(h->addr),
	      now,
	      h->xmit,
              h->returned,
	      h->best,
	      (int)((h->returned == 0) ? 0: h->total / h->returned),
	      h->worst);
	    else 
             fprintf(f,"%s\t%u\t%u\t%u\t%u\t%u\n",
	      addr_to_str(h->addr),
	      h->xmit,
              h->returned,
	      h->best,
	      (int)((h->returned == 0) ? 0: h->total / h->returned),
	      h->worst);
	    h->total=h->xmit=h->best=h->worst=h->returned=0;
	    h=h->next;
	    elements++;
        }
     }
  }
  fclose(f);
  output("%d routers in %d seconds",elements,time(NULL)-rtime);
  action("dump-routers",buffer);
}

/* Find next path to explore. Returns zero if a real ping was done. */
int next_ping() {
  int unknown=0;
  int i;
  struct nethost *h;
  if (hpaths==0) fatal("next_ping: called without any nethosts defined\n");

  /* Find next host to handle */
  if (cpath>=hpaths) cpath=0;
  current=hpath[cpath++];
  if (current==NULL || current->host->addr==0)
    fatal("Current path pointer is pointing nowhere (%d)!\n",cpath);
  for (i=0;i<current->n;i++) {
    if (current->path[i]==NULL || current->path[i]->addr == 0) unknown++;
    if (current->path[i] == current->host)
      unknown = 100; /* Make sure we drop into "we should restart" */
  }
  h=current->path[current->n];
  if ((h == current->host) ||
      (unknown > MAX_UNKNOWN_HOSTS) ||
      (current->n >= MaxPathLength-2)) {
    current->counter++;
    current->n = 0;
    if (Verbose) {
      output("Completed Path to %s routers=%d hosts=%d",addr_to_str(current->host->addr),hosts,hpaths);
    }
    /*  Check here for a routing change. Note that this might be triggered again and again since
        there is no marker that the script already ran */
/*    if (current->oldpath) action("route-change",addr_to_str(current->host->addr)); */
    return 1;
  } else { 
    if (h && h->time+FloodDelay > now()) { /* Make sure we do not hit one host too frequently */
	current->n++;	/* Assume that hop was fine if old data exists */
	return 1;
    }
    net_send_query(current->path[current->n]);
    current->n++;
    return 0;
  }
}


/*   We got a return on something we sent out.  Record the address and
     time.  */
void net_process_ping(int seq, uint32 addr, double now,int hop) {
  int totusec;
  int routechange=0;
  struct nethost *h;

  if (Verbose) output("Incoming from %s SEQ=%d Hop=%d",addr_to_str(addr),seq,hop);
  if(seq < 0 || seq >= MaxSequence) return;

  if(!sequence[seq].transit) return;

  sequence[seq].transit = 0;
  
  if (sequence[seq].h) {
      h=sequence[seq].h;
      if (h->addr != addr) {
          h=find_host(addr);
          if (h==NULL) h=add_host(addr);
          h->xmit++;
          routechange=1;
      }
  } else{
    h=find_host(addr);
    if (h==NULL)  h=add_host(addr);
    h->xmit++;
    /* Setup the path element link */
    sequence[seq].p->path[hop]=h;
  }
  totusec = (now-sequence[seq].time)*1000000;
  if(h->addr == 0) fatal("net_process_ping: zero host address in host structure\n");
  if(h->returned <= 0) {
    h->best = h->worst = totusec;
  }
  if(totusec < h->best) h->best = totusec;
  if(totusec > h->worst) h->worst = totusec;

  h->total += totusec;
  h->returned++;
  if (Verbose) { char *x;char *y;
     output("Route to %s hop %d is router %s",x=strdup(addr_to_str(sequence[seq].p->host->addr)),sequence[seq].p->n,y=strdup(addr_to_str(addr)));
     free(x);free(y);
  }
  if (routechange) { /* New route coming up*/
     struct netpath *x=sequence[seq].p;
     if (x->counter) { /* Do we have an already established route or is there only some instability ?? */
        /* We need to create an alternate route.. Look through already existing alternate routes for a match */
        struct netpath*y=x;
        while (y->oldpath && y->oldpath->path[hop]!=h) y=y->oldpath;
        if (y->oldpath) { /* Another existing path would fit here. Switch primary path */
          /* Note that this is just doing a guess. Routes could be confused now. */
          struct netpath scratch,*save;
          memcpy(&scratch,x,sizeof(struct netpath));
          y=y->oldpath;
          save=x->oldpath;memcpy(x,y,sizeof(struct netpath));x->oldpath=save;
          save=y->oldpath;memcpy(y,&scratch,sizeof(struct netpath));y->oldpath=save;
        } else {
            /* Completely new route */
	    y=malloc(sizeof(struct netpath));
            if (y==NULL) fatal("net_process_ping: cannot calloc netpath\n");
            memcpy(y,x,sizeof(struct netpath));
            x->oldpath=y;
            x->counter=0;
            bzero(x->path,MaxPathLength*sizeof(struct nethost *));
        }
     }
     x->path[hop]=h;
     x->n=0;		/* Verify route information from scratch */
  }
}

/*  We know a packet has come in, because the main select loop has called us,
    now we just need to read it, see if it is for us, and if it is a reply 
    to something we sent, then call net_process_ping()  */
void net_process_return() {
  char packet[2048];
  struct sockaddr_in fromaddr;
  int fromaddrsize;
  int num;
  struct ICMPHeader *header;
  double tnow;

  tnow=now();

  fromaddrsize = sizeof(fromaddr);
  num = recvfrom(recvsock, packet, 2048, 0, 
		 (struct sockaddr *)&fromaddr, &fromaddrsize);
  if (num<0) { output("net_process_return: recvfrom would have blocked!!");return; };

  if(num < sizeof(struct IPHeader) + sizeof(struct ICMPHeader))
    return;

  header = (struct ICMPHeader *)(packet + sizeof(struct IPHeader));
  if(header->type == ICMP_ECHOREPLY) {
    if(header->id < pid || header->id >= pid+MaxPathLength)
      return;

    net_process_ping(header->sequence, fromaddr.sin_addr.s_addr, tnow,header->id-pid);
  } else if(header->type == ICMP_TIME_EXCEEDED) {
    if(num < sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + 
             sizeof(struct IPHeader) + sizeof(struct ICMPHeader))
      return;
    
    header = (struct ICMPHeader *)(packet + sizeof(struct IPHeader) + 
				sizeof(struct ICMPHeader) + sizeof(struct IPHeader));
    if(header->id < pid || header->id >=pid+MaxPathLength)
      return;

    net_process_ping(header->sequence, fromaddr.sin_addr.s_addr, tnow,header->id-pid);
  }
}


void net_reset() {
  int at;

  for(at = 0; at < MaxHost; at++) if (host[at]) {
    host[at]->xmit = 0;
    host[at]->returned = 0;
    host[at]->total = 0;
    host[at]->best = 0;
    host[at]->worst = 0;
  }
  
  for(at = 0; at < MaxSequence; at++) {
    sequence[at].transit = 0;
  }

}

void net_close() {
  close(sendsock);
  close(recvsock);
}

int net_waitfd() {
  return recvsock;
}


     