/*
    pfd -- path finder daemon  --  a network diagnostic tool
    Copyright (C) 1997,1998  Matt Kimball
    Copyright (C) 2000,2001  Christoph Lameter

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include "getopt.h"
#include "net.h"
#include "output.h"

#ifndef HAVE_SETEUID
/* HPUX doesn't have seteuid, but setuid works fine in that case for us */
#define seteuid setuid
#endif

int Verbose = 0;
int format=1;
int PrintHelp = 0;
char *datapath=NULL;
char *HostList =  NULL;
char *actionscript = NULL;

int packetsize = MINPACKET;
int rate = 10;	/* Number of pings per second */
int floodrate =10;	/* Maximum number of pings to one host per second */
int daemonmode=0;		/* Keep on running or terminate after first run */
double WriteInterval = 300;
int terminate = 0;	/* Set to one to terminate */

/* Different timers */
double dumptime,pingtime;
double PingInterval;
double FloodDelay;


void parse_arg(int argc, char **argv) {
  int opt;
  static struct option long_options[] = {
    { "verbose", 0, 0, 'v' },
    { "help", 0, 0, 'h' },
    { "rate", 1, 0, 'r' },
    { "write-interval",1, 0, 'w' },
    { "floodrate",1,0,'f'},
    { "psize", 1, 0, 'p' },
    { "daemon",0, 0, 'd' },
    { "action",1,0,'a' },
    { "mysql",0,0, 'm' },
    { "outputpath",1,0,'o' },
    { 0, 0, 0, 0 }
  };

  opt = 0;
  while(1) {
    opt = getopt_long(argc, argv, "a:dhmvp:w:do:", long_options, NULL);
    if(opt == -1)
      break;

    switch(opt) {
    case 'v':
      Verbose = 1;
      break;
    case 'd':
      daemonmode=1;
      break;
    case 'h':
      PrintHelp = 1;
      break;
    case 'f':
      floodrate= atoi(optarg);
      break;
    case 'm':
      format=0;
      break;
    case 'w':
      WriteInterval = atof(optarg)*60;
      break;
    case 'p':
      packetsize = atoi (optarg);
      break;
    case 'r':
      rate=atoi(optarg);
      if (rate <= 0) {
	fprintf (stderr, "pfd: rate must be positive\n");
	exit (1);
      }
      break;
    case 'a':
      actionscript=strdup(optarg);
      break;
    case 'o':
      datapath=strdup(optarg);
    }
  }

  if(optind > argc - 1)
    return;

  HostList = argv[optind++];

  if (argc > optind) 
    packetsize = atoi(argv[optind]);

}


extern int hpaths;

void set_termination_flag(int signum) {
	output("set_termination_flag: signal %d\n");
	terminate=1;
}

void handle_child(int signum) {
    int status,pid;
 	pid=wait(&status);
    	output("handle_child: signal %d, pid %d",signum,pid);
}

/* Execute an action script */
void action(char *token,char *data) {
    if (actionscript) {
	if (fork()==0) {
	  execlp(actionscript,actionscript,token,data,NULL);
	  perror(actionscript);
	  exit(1);
      	}
    }
}

int main(int argc, char **argv) {
  int net_preopen_result;
  fd_set readfd;
  int anyset;
  int netfd;
  struct timeval selecttime;

  parse_arg(argc, argv);

  if(PrintHelp) {
    printf("usage: %s [-hvrmnip] [--help] [--verbose]\n"
           "\t\t[--psize=bytes/-p=bytes]\n"            /* ok */
           "\t\t[--write-interval=MINUTES/-w=MINUTES]\n"
           "\t\t[--floodrate=number-per-second/-f=SECONDS]\n"
           "\t\t[--daemon]/-d\n"
           "\t\t[--action=SCRIPT/-a=SCRIPT]\n"
           "\t\t[--mysql/-m]\n"
           "\t\t[--outputpath=DIRECTORY/-o=DIRECTORY]\n"
	   "\t\t[--rate/-r=number-per-second [FILE [PACKETSIZE]]\n", argv[0]);
    exit(0);
  }

  if (daemonmode) {
	if (HostList==NULL) HostList="/etc/pathfinder.list";
	if (fork()) exit(0);
	output_open();
	if (datapath==NULL) datapath="/var/log/pathfinder";
  }

  net_preopen_result = net_init ();

  if (!daemonmode) {
      /*  Now drop to user permissions  */
      if(setuid(getuid())) {
	printf("pfd: Unable to drop permissions.\n");
        exit(1);
      }
      /*  Double check, just in case  */
      if(geteuid() != getuid()) {
        printf("pfd: Unable to drop permissions.\n");
        exit(1);
      }
      if (datapath==NULL) datapath=".";
  }

  if(net_preopen_result != 0) {
    output("Unable to get raw socket.");
    exit(1);
  };
  
  { FILE *f;char x[200],*p;struct hostent *h;
    if (HostList) {
      output("Reading Hostlist from %s",HostList);
      if ((f=fopen(HostList,"r"))==NULL) {
        output("Cannot open hostlist in %s", HostList);
        exit(1);
      }
    } else f=stdin;
    while (fgets(x,sizeof(x)-1,f)) {
        p=x;
        while (isspace(*p)) p++;
        if (*p=='#'||*p==0||*p=='\n') continue;
        p[strlen(p)-1]=0;
  	h=gethostbyname(p);
  	if (h) {
  	    net_add_netpath(*(int *)h->h_addr);
  	    if (Verbose) output("Host=%s IP=%s added",x,addr_to_str(*(int *)h->h_addr));
  	} else
  	    output("Cannot resolve  hostname %s",p);
    }
    fclose(f);
  }
  net_reset();

  PingInterval=(double)1/(double)rate;
  FloodDelay=(double)1/(double)floodrate;

  pingtime=now();
  dumptime=pingtime+WriteInterval;
  output("Starting to monitor %d paths. Ping Interval=%.2fms Flood Delay=%.2fms Data Dump=%.2fmin",hpaths,PingInterval*1000,FloodDelay*1000,WriteInterval/60);

  if (signal(SIGINT,set_termination_flag) == SIG_IGN) signal(SIGINT,SIG_IGN);
  if (signal(SIGHUP,set_termination_flag) == SIG_IGN) signal(SIGHUP,SIG_IGN);
  if (signal(SIGTERM,set_termination_flag) == SIG_IGN) signal(SIGTERM,SIG_IGN);
  signal(SIGCHLD,handle_child);
  anyset = 0;
  while(!terminate) {


    selecttime.tv_sec = 0;
    selecttime.tv_usec = 0;

    if(!anyset) { double x,y;
      y=now();
      /* Nothing urgent to do. Check for things due to be done */
      if (y>pingtime && pingtime <dumptime-30) {
	pingtime+=PingInterval;
	while (next_ping());
      } else
      if (y>dumptime) { dumptime+=WriteInterval;
        if (daemonmode) data_dump(); else terminate=1;
        /* The abov will take awhile so avoid flooding by resetting ping clock */
        pingtime=now();
      }
      /* When is the most likely next event ?*/
      x=dumptime;
      if (pingtime <x) x=pingtime;
      x-=y;
      if (x>0) {
        selecttime.tv_sec = floor(x);
        selecttime.tv_usec=(x-selecttime.tv_sec)*1000000;
      }
    }
    FD_ZERO(&readfd);
    netfd = net_waitfd();
    FD_SET(netfd, &readfd);
    anyset=0;
    if (select(netfd+1, (void *)&readfd, NULL, NULL, &selecttime)>0) {
	/*  Have we got new packets back?  */
	if(FD_ISSET(netfd, &readfd)) {
	   net_process_return();
	   anyset = 1;
	}
    }
  }
  net_close();
  data_dump();
  return 0;
}
